package validator

import (
	"regexp"
	"slices"
	"strings"
	"time"
	"unicode/utf8"
)

type Validator struct {
	NonFieldErrors []string
	FieldErrors    map[string]string
}

func (v *Validator) Valid() bool {
	return len(v.FieldErrors) == 0 && len(v.NonFieldErrors) == 0
}

func (v *Validator) AddNonFieldError(message string) {
	v.NonFieldErrors = append(v.NonFieldErrors, message)
}

func (v *Validator) AddFieldError(key, message string) {
	if v.FieldErrors == nil {
		v.FieldErrors = make(map[string]string)
	}
	if _, exists := v.FieldErrors[key]; !exists {
		v.FieldErrors[key] = message
	}
}

func (v *Validator) CheckField(ok bool, key, message string) {
	if !ok {
		v.AddFieldError(key, message)
	}
}

func Blank(value string) bool {
	return strings.TrimSpace(value) == ""
}

func NotBlank(value string) bool {
	return strings.TrimSpace(value) != ""
}

func MinChars(value string, n int) bool {
	return utf8.RuneCountInString(value) >= n
}

func MaxChars(value string, n int) bool {
	return utf8.RuneCountInString(value) <= n
}

func PermittedValue[T comparable](value T, permittedValues ...T) bool {
	return slices.Contains(permittedValues, value)
}

func Matches(value string, rx *regexp.Regexp) bool {
	return rx.MatchString(value)
}

func GreaterThan(value int, n int) bool {
	return value > n
}

func LesserThan(value int, n int) bool {
	return value < n
}

func GreaterThanOrEquals(value int, n int) bool {
	return value >= n
}

func After(value time.Time, n time.Time) bool {
	return value.After(n)
}

func UniqueValues[T comparable](values []T) bool {
	seen := make(map[T]bool)

	for _, value := range values {
		if seen[value] {
			return false
		}
		seen[value] = true
	}

	return true
}
